<?php

namespace Drupal\commerce_payment_extra_order;

use Drupal\commerce_order\Entity\OrderInterface;
use Drupal\commerce_payment_extra\PaymentManager;
use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Psr\Log\LoggerInterface;

class OrderAuthorizedManager {

  /**
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  protected $config;

  /**
   * @var \Drupal\Component\Datetime\TimeInterface
   */
  protected $time;

  /**
   * @var \Drupal\commerce_payment_extra\PaymentManager
   */
  protected $paymentManager;

  /**
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * OrderAuthorizedManager constructor.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $configFactory
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   * @param \Drupal\Component\Datetime\TimeInterface $time
   * @param \Drupal\commerce_payment_extra\PaymentManager $paymentManager
   * @param \Psr\Log\LoggerInterface $logger
   */
  public function __construct(ConfigFactoryInterface $configFactory, EntityTypeManagerInterface $entityTypeManager, TimeInterface $time, PaymentManager $paymentManager, LoggerInterface $logger) {
    $this->config = $configFactory->get('commerce_payment_extra_order.settings');
    $this->entityTypeManager = $entityTypeManager;
    $this->time = $time;
    $this->paymentManager = $paymentManager;
    $this->logger = $logger;
  }

  public function processOrders() {
    $storage = $this->entityTypeManager->getStorage('commerce_order');
    $order_ids = $this->getOrders();
    if (empty($order_ids)) {
      $this->logger->debug('No orders found matching fulfilment criteria.');
    }
    foreach ($order_ids as $order_id) {
      $order = $storage->loadUnchanged($order_id);
      try {
        $this->processOrder($order);
      }
      catch (\Throwable $e) {
        $this->logger->error('Caught exception trying to process order id @order_id: @msg. Stack trace: @trace', [
          '@msg' => $e->getMessage(),
          '@trace' => $e->getTraceAsString(),
          '@order_id' => $order_id,
        ]);
      }
    }
  }

  /**
   * @return array
   */
  protected function getOrders() {
    $storage = $this->entityTypeManager->getStorage('commerce_order');
    $query = $storage->getQuery();

    // @todo: Only order with payments are applicable. This query should join
    // payment table.

    $query->accessCheck(FALSE);
    $query->condition('state', 'draft');
    $query->condition('locked', 1);
    // By default process only orders between 1 h and 30 days since last update.
    $query->condition('changed', $this->time->getRequestTime() - ($this->config->get('authorized_auto_place_min_threshold') ?: 3600), '<');
    $query->condition('changed', $this->time->getRequestTime() - ($this->config->get('authorized_auto_place_max_threshold') ?: 2592000), '>');
    return $query->execute();
  }

  /**
   * @param \Drupal\commerce_order\Entity\OrderInterface $order
   */
  protected function processOrder(OrderInterface $order) {
    if ($order->getState()->getId() != 'draft') {
      $this->logger->warning('The order @order_id has already been placed', [
        '@order_id' => $order->id(),
      ]);
      return;
    }
    /** @var \Drupal\commerce_payment\Entity\PaymentGatewayInterface $payment_gateway */
    $payment_gateway = $order->get('payment_gateway')->entity;
    if (!$payment_gateway) {
      $this->logger->emergency('The payment gateway is unknown for order @order_id', [
        '@order_id' => $order->id(),
      ]);
      return;
    }

    $balance = $order->getBalance();
    foreach ($payments = $this->paymentManager->loadCapturablePaymentsByOrder($order) as $payment) {
      if (!in_array($payment->getPaymentGatewayId(), $this->config->get('supported_payment_gateways') ?: [])) {
        continue;
      }
      $balance = $balance->subtract($payment->getAmount());
    }
    if (empty($payments)) {
      $this->logger->debug('Order @order_id is locked and in draft state but no payments were found to cover it.', [
        '@order_id' => $order->id(),
      ]);
      return;
    }
    elseif ($balance->isPositive()) {
      $this->logger->emergency('Insufficient funds for order @order_id. Uncovered balance @balance.', [
        '@order_id' => $order->id(),
        '@balance' => $balance->getNumber(),
      ]);
      return;
    }

    $order->getState()->applyTransitionById('place');
    // A placed order should never be locked.
    $order->unlock();
    $order->save();
    $this->logger->notice('The order @order_id has been placed automatically', [
      '@order_id' => $order->id(),
    ]);
  }
}
